@@ -23,7 +23,7 @@ Allocation conventions:
 Return type conventions:
 
  - A return type `rktio_ok_t` (alias for `int`) means that 1 is
-   returned for succes and 0 for error. Use
+   returned for success and 0 for error. Use
    `rktio_get_last_error_kind` and `rktio_get_last_error` for more
    information about a 0 result.
 
@@ -63,7 +63,7 @@ Thread and signal conventions:
    before a second call, different `rktio_t` values can be used freely
    from different threads.
 
- - Unless otherwise specificed, anything created with a particular
+ - Unless otherwise specified, anything created with a particular
    `rktio_t` must be used with that same `rktio_t` thereafter (and in
    only one thread at a time).
 
@@ -76,7 +76,7 @@ Thread and signal conventions:
 
  - On systems where signal handling is thread-specific, as on Linux,
    then `rktio_init` should be called before any additional threads,
-   so that a suitable inheritable siganl disposition can be
+   so that a suitable inheritable signal disposition can be
    configured.
 
 */
@@ -91,8 +91,9 @@ Thread and signal conventions:
 #define RKTIO_EXTERN_NOERR  RKTIO_EXTERN
 #define RKTIO_EXTERN_STEP   RKTIO_EXTERN
 
-#define RKTIO_NULLABLE      /* empty */
-#define RKTIO_BLOCKING      /* empty */
+#define RKTIO_NULLABLE      /* empty; pointer type can be NULL */
+#define RKTIO_BLOCKING      /* empty; function blocks indefinitely */
+#define RKTIO_MSG_QUEUE     /* empty; function can dispatch events on Windows */
 
 /*************************************************/
 /* Initialization and general datatypes          */
@@ -147,7 +148,7 @@ RKTIO_EXTERN void rktio_set_dll_path(rktio_char16_t *p);
    specific to a `rktio_t` instance. */
 
 RKTIO_EXTERN rktio_char16_t *rktio_get_dll_path(rktio_char16_t *p);
-/* Combines a path prevously registered with `rktio_set_dll_path` with
+/* Combines a path previously registered with `rktio_set_dll_path` with
    the given filename. The result is allocated (as should be
    deallocated) as usual. */
 
@@ -176,7 +177,7 @@ typedef struct rktio_fd_t rktio_fd_t;
 #define RKTIO_OPEN_UDP         (1<<8)
 #define RKTIO_OPEN_REGFILE     (1<<9)
 #define RKTIO_OPEN_NOT_REGFILE (1<<10)
-/* If neither RKTIO_OPEN_REGILE nor RKTIO_OPEN_NOT_REGILE
+/* If neither RKTIO_OPEN_REGFILE nor RKTIO_OPEN_NOT_REGFILE
    are specified, then the value is inferred by `rtkio_system_fd`. */
 #define RKTIO_OPEN_DIR         (1<<11)
 #define RKTIO_OPEN_NOT_DIR     (1<<12)
@@ -187,6 +188,9 @@ typedef struct rktio_fd_t rktio_fd_t;
 #define RKTIO_OPEN_OWN         (1<<14)
 /* Make `rtkio_system_fd` record a socket for reliable clean up on pre-NT Windows. */
 
+/* Used for `rktio_open` with `RKTIO_OPEN_WRITE`: */
+#define RKTIO_OPEN_REPLACE_PERMS (1<<15)
+
 RKTIO_EXTERN rktio_fd_t *rktio_system_fd(rktio_t *rktio, intptr_t system_fd, int modes);
 /* A socket (as opposed to other file descriptors) registered this way
    should include include `RKTIO_OPEN_SOCKET` and be non-blocking or
@@ -227,6 +231,16 @@ RKTIO_EXTERN rktio_fd_t *rktio_open(rktio_t *rktio, rktio_const_string_t src, in
    without `RKTIO_OPEN_READ`, then the result may be a file descriptor
    in pending-open mode until the read end is opened. */
 
+RKTIO_EXTERN rktio_fd_t *rktio_open_with_create_permissions(rktio_t *rktio,
+                                                            rktio_const_string_t src,
+                                                            int modes, int perm_bits);
+/* Like `rktio_open`, but accepts permission bits that are used if the
+   file is created (which is only relevant if `modes` includes
+   `RKTIO_OPEN_WRITE`). On Unix, perm_bits are adjusted by a umask.
+   Otherwise, permission bits are treated in the same way as
+   by `rktio_set_file_or_directory_permissions`. */
+#define RKTIO_DEFAULT_PERM_BITS 0666
+
 RKTIO_EXTERN rktio_ok_t rktio_close(rktio_t *rktio, rktio_fd_t *fd);
 /* Can report `RKTIO_ERROR_EXISTS` in place of system error,
    and can report `RKTIO_ERROR_UNSUPPORTED_TEXT_MODE` on Windows.
@@ -234,13 +248,13 @@ RKTIO_EXTERN rktio_ok_t rktio_close(rktio_t *rktio, rktio_fd_t *fd);
 
 RKTIO_EXTERN void rktio_close_noerr(rktio_t *rktio, rktio_fd_t *fd);
 /* The same as `rktio_close`, but without reporting errors. There's
-   often nothing good to do if a close fails, epsecially if the close
+   often nothing good to do if a close fails, especially if the close
    is in the service of handling another failure where you don't want
    the error code replaced. */
 
 RKTIO_EXTERN rktio_fd_t *rktio_dup(rktio_t *rktio, rktio_fd_t *rfd);
 /* Copies a file descriptor, where each must be closed or forgotten
-   independenty. */
+   independently. */
 
 RKTIO_EXTERN void rktio_forget(rktio_t *rktio, rktio_fd_t *fd);
 /* Deallocates a `rktio_fd_t` without closing the file descriptor,
@@ -327,7 +341,7 @@ rktio_tri_t rktio_file_lock_try(rktio_t *rktio, rktio_fd_t *rfd, rktio_bool_t ex
 RKTIO_EXTERN rktio_ok_t rktio_file_unlock(rktio_t *rktio, rktio_fd_t *rfd);
 /* Advisory file locks, where `excl` attempts to claim an exclusive
    lock. Whether these work in various situations depend on many OS
-   details, where the differences involve promoting from non-exlcusive
+   details, where the differences involve promoting from non-exclusive
    to exclusive, taking a lock that is already held, getting an
    exclusive lock for a file descriptor in read mode, getting a
    non-exclusive lock in write mode, and whether a lock prevents
@@ -361,7 +375,7 @@ typedef struct rktio_fd_transfer_t rktio_fd_transfer_t;
 
 RKTIO_EXTERN_NOERR rktio_fd_transfer_t *rktio_fd_detach(rktio_t *rktio, rktio_fd_t *rfd);
 /* Returns a variant of `rfd` that does not depend on `rktio`. The
-   `rfd` must not currently have any file locks, and deatching
+   `rfd` must not currently have any file locks, and detaching
    transfers ownership of `rfd` to the result. To use the result, it
    must be reattached to some `rktio_t` using rktio_fd_attach, or it
    can be freed with `rktio_fd_free_transfer`. */
@@ -381,7 +395,8 @@ RKTIO_EXTERN rktio_fd_t **rktio_make_pipe(rktio_t *rktio, int flags);
 /* Makes a pair of file descriptors for a pipe. The first one
    is the read end, and the second is the write end. The `flags`
    can declare the intended sharing of the file descriptors
-   with a child process, and is useful only on Windows. */
+   with a child process on Windows and platforms where O_CLOEXEC
+   is used. */
 /* For `flags`: */
 #define RKTIO_NO_INHERIT_INPUT  (1<<0)
 #define RKTIO_NO_INHERIT_OUTPUT (1<<1)
@@ -433,7 +448,7 @@ RKTIO_EXTERN rktio_fd_t *rktio_accept(rktio_t *rktio, rktio_listener_t *listener
 RKTIO_EXTERN rktio_connect_t *rktio_start_connect(rktio_t *rktio,
                                                   rktio_addrinfo_t *remote,
                                                   RKTIO_NULLABLE rktio_addrinfo_t *local);
-/* Starts a connection request. Addreses must not be freed until the
+/* Starts a connection request. Addresses must not be freed until the
    connection is complete, errored, or stopped. */
 
 RKTIO_EXTERN rktio_fd_t *rktio_connect_finish(rktio_t *rktio, rktio_connect_t *conn);
@@ -459,11 +474,15 @@ RKTIO_EXTERN rktio_fd_t *rktio_connect_trying(rktio_t *rktio, rktio_connect_t *c
 
 RKTIO_EXTERN rktio_ok_t rktio_socket_shutdown(rktio_t *rktio, rktio_fd_t *rfd, int mode);
 /* Useful for TCP to report an EOF to the other end. Does not close the socket,
-   but may make it ineligible for forther use.
+   but may make it ineligible for further use.
    `mode` values: */
 #define RKTIO_SHUTDOWN_READ   0
 #define RKTIO_SHUTDOWN_WRITE  1
 
+RKTIO_EXTERN rktio_ok_t rktio_tcp_nodelay(rktio_t *rktio, rktio_fd_t *rfd, rktio_bool_t enable);
+/* Changes a connection to enable or disable "TCP_NODELAY" mode,
+   which diabled Nagle's algorithm for avoiding small packets. */
+
 RKTIO_EXTERN rktio_fd_t *rktio_udp_open(rktio_t *rktio, RKTIO_NULLABLE rktio_addrinfo_t *addr, int family);
 /* The `addr` argument can be NULL to create a socket without
    specifying an interface, and `family` is used only if `addr` is not
@@ -605,6 +624,8 @@ RKTIO_EXTERN rktio_process_result_t *rktio_process(rktio_t *rktio,
 #define RKTIO_PROCESS_STDOUT_AS_STDERR          (1<<1)
 #define RKTIO_PROCESS_WINDOWS_EXACT_CMDLINE     (1<<2)
 #define RKTIO_PROCESS_WINDOWS_CHAIN_TERMINATION (1<<3)
+#define RKTIO_PROCESS_NO_CLOSE_FDS              (1<<4)
+#define RKTIO_PROCESS_NO_INHERIT_FDS            (1<<5)
 
 RKTIO_EXTERN_NOERR int rktio_process_allowed_flags(rktio_t *rktio);
 /* Reports the flags that are accepted by `rktio_process` on the
@@ -659,7 +680,7 @@ struct rktio_ltps_t; /* forward reference */
 RKTIO_EXTERN rktio_fs_change_t *rktio_fs_change(rktio_t *rktio, rktio_const_string_t path,
                                                 struct rktio_ltps_t *ltps);
 /* Creates a filesystem-change tracker that reports changes in `path`
-   after creation of the tracker. The properties repotred by
+   after creation of the tracker. The properties reported by
    `rktio_fs_change_properties` report various aspects of how the
    tracker behaves. In particular, the `ltps` argument can be NULL
    unless the `RKTIO_FS_CHANGE_NEED_LTPS` property is reported; if
@@ -724,7 +745,7 @@ RKTIO_EXTERN void rkio_reset_sleep_backoff(rktio_t *rktio);
 
 /* "Long-term" means that the poll set will be used frequently with
    incremental updates, which means that it's worthwhile to use an OS
-   facililty (epoll, kqueue, etc.) to speed up polling. */
+   facility (epoll, kqueue, etc.) to speed up polling. */
 
 typedef struct rktio_ltps_t rktio_ltps_t;
 typedef struct rktio_ltps_handle_t rktio_ltps_handle_t;
@@ -732,7 +753,7 @@ typedef struct rktio_ltps_handle_t rktio_ltps_handle_t;
 RKTIO_EXTERN rktio_ltps_t *rktio_ltps_open(rktio_t *rktio);
 
 RKTIO_EXTERN void rktio_ltps_close(rktio_t *rktio, rktio_ltps_t *lt);
-/* Closing will signal all remianing handles and free all signaled
+/* Closing will signal all remaining handles and free all signaled
    handles, but use `rktio_ltps_remove_all` and
    `rktio_ltps_get_signaled_handle` is you need to clean up any
    per-handle data: */
@@ -765,7 +786,7 @@ RKTIO_EXTERN void rktio_ltps_remove_all(rktio_t *rktio, rktio_ltps_t *lt);
 /* Removes all additions, signaling all handles. */
 
 RKTIO_EXTERN rktio_ok_t rktio_ltps_poll(rktio_t *rktio, rktio_ltps_t *lt);
-/* Enqueues signaled handles for retreival via `rktio_ltps_get_signaled_handle`.  */
+/* Enqueues signaled handles for retrieval via `rktio_ltps_get_signaled_handle`.  */
 
 RKTIO_EXTERN rktio_ltps_handle_t *rktio_ltps_get_signaled_handle(rktio_t *rktio, rktio_ltps_t *lt);
 /* Free the returned handle when you're done with it. */
@@ -773,7 +794,7 @@ RKTIO_EXTERN rktio_ltps_handle_t *rktio_ltps_get_signaled_handle(rktio_t *rktio,
 RKTIO_EXTERN void rktio_ltps_handle_set_auto(rktio_t *rktio, rktio_ltps_handle_t *lth, int auto_mode);
 /* An alternative to receiving the handle via `rktio_ltps_get_signaled_handle`;
    have signaling automatically either zero the handle content (so the
-   client can detect signaling) or free the handle (bcause the client
+   client can detect signaling) or free the handle (because the client
    is no longer watching it). If `auto_mode` is `RKTIO_LTPS_HANDLE_NONE`,
    automatic handling is disabled for the handle. */
 /* `auto_mode` values: */
@@ -836,6 +857,10 @@ RKTIO_EXTERN rktio_ok_t rktio_set_current_directory(rktio_t *rktio, rktio_const_
 RKTIO_EXTERN rktio_ok_t rktio_make_directory(rktio_t *rktio, rktio_const_string_t filename);
 /* Can report `RKTIO_ERROR_EXISTS`. */
 
+RKTIO_EXTERN rktio_ok_t rktio_make_directory_with_permissions(rktio_t *rktio, rktio_const_string_t filename, int perm_bits);
+/* Can report `RKTIO_ERROR_EXISTS`. */
+#define RKTIO_DEFAULT_DIRECTORY_PERM_BITS 0777
+
 RKTIO_EXTERN rktio_ok_t rktio_delete_directory(rktio_t *rktio, rktio_const_string_t filename, rktio_const_string_t current_directory,
                                                rktio_bool_t enable_write_on_fail);
 /* The `current_directory` argument is used on Windows to avoid being
@@ -861,6 +886,21 @@ RKTIO_EXTERN rktio_filesize_t *rktio_file_size(rktio_t *rktio, rktio_const_strin
 RKTIO_EXTERN rktio_timestamp_t *rktio_get_file_modify_seconds(rktio_t *rktio, rktio_const_string_t file);
 RKTIO_EXTERN rktio_ok_t rktio_set_file_modify_seconds(rktio_t *rktio, rktio_const_string_t file, rktio_timestamp_t secs);
 
+typedef struct rktio_stat_t {
+  /* Eventually, this should use `int64_t`, available in C99 and up */
+  uintptr_t device_id, inode, mode, hardlink_count, user_id, group_id,
+            device_id_for_special_file, size, block_size, block_count,
+            access_time_seconds, access_time_nanoseconds,
+            modify_time_seconds, modify_time_nanoseconds,
+            ctime_seconds, ctime_nanoseconds;
+  /* The `st_ctime` field is status change time for Posix and creation time
+     for Windows. */
+  rktio_bool_t ctime_is_change_time;
+} rktio_stat_t;
+
+RKTIO_EXTERN rktio_stat_t *rktio_file_or_directory_stat(rktio_t *rktio, rktio_const_string_t path, rktio_bool_t follow_links);
+RKTIO_EXTERN rktio_stat_t *rktio_fd_stat(rktio_t *rktio, rktio_fd_t *fd);
+
 typedef struct rktio_identity_t {
   uintptr_t a, b, c;
   int a_bits, b_bits, c_bits; /* size of each in bits */
@@ -921,6 +961,17 @@ RKTIO_EXTERN_STEP rktio_file_copy_t *rktio_copy_file_start(rktio_t *rktio, rktio
    whole copy, or it may just get started. Can report
    `RKTIO_ERROR_EXISTS`, and sets an error step as listed further below. */
 
+RKTIO_EXTERN_STEP rktio_file_copy_t *rktio_copy_file_start_permissions(rktio_t *rktio, rktio_const_string_t dest, rktio_const_string_t src,
+                                                                       rktio_bool_t exists_ok,
+                                                                       rktio_bool_t use_perm_bits, int perm_bits,
+                                                                       rktio_bool_t override_create_perms);
+/* Like `rktio_copy_file_start`, but accepts optional permissions to
+   apply to the copy, which are used only if `use_perm_bits` is set
+   (otherwise the source file's permissions are kept) and whether on
+   Unix to for the destination file's permissions as possibly modified
+   by `umask` on file create (whether supplied or taken from the `src`
+   file) or because the file already exists. */
+
 RKTIO_EXTERN rktio_bool_t rktio_copy_file_is_done(rktio_t *rktio, rktio_file_copy_t *fc);
 RKTIO_EXTERN_STEP rktio_ok_t rktio_copy_file_step(rktio_t *rktio, rktio_file_copy_t *fc);
 /* As long as the copy isn't done, call `rktio_copy_file_step` to make
@@ -1052,16 +1103,21 @@ typedef struct rktio_date_t {
 } rktio_date_t;
 
 RKTIO_EXTERN_NOERR uintptr_t rktio_get_milliseconds(void);
-/* Overflow may cause the result to wrap around to 0, at least on a
-   32-bit platform. */
+/* Wll-clock time. Overflow may cause the result to wrap around to 0,
+   at least on a 32-bit platform. */
 
 RKTIO_EXTERN_NOERR double rktio_get_inexact_milliseconds(void);
-/* No overflow, but won't strictly increase if the system clock is reset. */
+/* Wall-clock time. No overflow, but won't strictly increase if the
+   system clock is reset. */
+
+RKTIO_EXTERN_NOERR double rktio_get_inexact_monotonic_milliseconds(rktio_t *rktio);
+/* Real time like wall-clock time, but will strictly increase,
+   assuming that the host system provides a monotonic clock. */
 
 RKTIO_EXTERN_NOERR uintptr_t rktio_get_process_milliseconds(rktio_t *rktio);
 RKTIO_EXTERN_NOERR uintptr_t rktio_get_process_children_milliseconds(rktio_t *rktio);
-/* Overflow may cause the result to wrap around to 0, at least on a
-   32-bit platform. */
+/* CPU time across all threads withing the process. Overflow may cause
+   the result to wrap around to 0, at least on a 32-bit platform. */
 
 RKTIO_EXTERN_NOERR rktio_timestamp_t rktio_get_seconds(rktio_t *rktio);
 RKTIO_EXTERN rktio_date_t *rktio_seconds_to_date(rktio_t *rktio, rktio_timestamp_t seconds, int nanoseconds, int get_gmt);
@@ -1085,12 +1141,12 @@ enum {
   RKTIO_SW_SHOWNORMAL
 };
 
-RKTIO_EXTERN rktio_ok_t rktio_shell_execute(rktio_t *rktio,
-                                            rktio_const_string_t verb,
-                                            rktio_const_string_t target,
-                                            rktio_const_string_t arg,
-                                            rktio_const_string_t dir,
-                                            int show_mode);
+RKTIO_EXTERN RKTIO_MSG_QUEUE rktio_ok_t rktio_shell_execute(rktio_t *rktio,
+                                                            rktio_const_string_t verb,
+                                                            rktio_const_string_t target,
+                                                            rktio_const_string_t arg,
+                                                            rktio_const_string_t dir,
+                                                            int show_mode);
 /* Supported only on Windows to run `ShellExecute`. The `dir` argument
    needs to have normalized path separators. */
 
@@ -1108,7 +1164,7 @@ RKTIO_EXTERN_NOERR char *rktio_wide_path_to_path(rktio_t *rktio, const rktio_cha
 
 RKTIO_EXTERN_NOERR int rktio_processor_count(rktio_t *rktio);
 /* Returns the number of processing units, either as CPUs, cores, or
-   hyoperthreads. */
+   hyperthreads. */
 
 /*************************************************/
 /* Logging                                       */
@@ -1157,10 +1213,10 @@ intptr_t rktio_convert(rktio_t *rktio,
    increments `*out` and decrements `*out_left`. In case of an error,
    the result is `RKTIO_CONVERT_ERROR` and the last error is set to
    one of `RKTIO_ERROR_CONVERT_NOT_ENOUGH_SPACE`,
-   `RKTIO_ERROR_CONVERT_BAD_SEQUENCE`, or `RKTIO_ERROR_CONVERT_OTHER`
-   --- but an error indicates something within `in` or `out`,
-   and some bytes may have been successfully converted even if an
-   error is reported. */
+   `RKTIO_ERROR_CONVERT_BAD_SEQUENCE`, `RKTIO_ERROR_CONVERT_PREMATURE_END`,
+   or `RKTIO_ERROR_CONVERT_OTHER` --- but an error indicates something within
+   `in` or `out`, and some bytes may have been successfully converted even if
+   an error is reported. */
 
 #define RKTIO_CONVERT_ERROR (-1)
 
@@ -1194,7 +1250,7 @@ RKTIO_EXTERN_NOERR rktio_char16_t *rktio_recase_utf16(rktio_t *rktio,
 /* Converts the case of a string encoded in UTF-16 for the system's
    default locale if the OS provided direct support for it. The
    `RKTIO_CONVERT_RECASE_UTF16 property from
-   `rktio_convert_properties` reports whether this functon will work.
+   `rktio_convert_properties` reports whether this function will work.
    Takes and optionally returns a length (`olen` can be NULL), but the
    UTF-16 sequence is expected to have no nuls. */
 
@@ -1210,7 +1266,7 @@ RKTIO_EXTERN_NOERR int rktio_strcoll_utf16(rktio_t *rktio,
 /* Compares two strings encoded in UTF-16 for the system's default
    locale if the OS provided direct support for it. The
    `RKTIO_CONVERT_STRCOLL_UTF16 property from
-   `rktio_convert_properties` reports whether this functon will work.
+   `rktio_convert_properties` reports whether this function will work.
    Takes lengths, but the UTF-16 sequences are expected to have
    no include nuls. */
 
@@ -1301,7 +1357,7 @@ RKTIO_EXTERN rktio_dll_t *rktio_dll_open(rktio_t *rktio, rktio_const_string_t na
    If a DLL has been loaded with `name` already, the previous result
    is returned again, but with an internal reference count returned.
    The `as_global` argument matters only for the first load of a DLL
-   thrrough a given `name`.
+   through a given `name`.
 
    Unless the DLL is explicitly unloaded with `rktio_dll_close`, even
    when the given `rktio` is closed with `rktio_destroy`, loaded
@@ -1368,7 +1424,7 @@ enum {
   RKTIO_ERROR_INIT_FAILED,
   RKTIO_ERROR_LTPS_NOT_FOUND,
   RKTIO_ERROR_LTPS_REMOVED, /* indicates success, instead of failure */
-  RKTIO_ERROR_CONNECT_TRYING_NEXT, /* indicates that failure is not (yet) premanent */
+  RKTIO_ERROR_CONNECT_TRYING_NEXT, /* indicates that failure is not (yet) permanent */
   RKTIO_ERROR_ACCEPT_NOT_READY,
   RKTIO_ERROR_HOST_AND_PORT_BOTH_UNSPECIFIED,
   RKTIO_ERROR_INFO_TRY_AGAIN, /* for UDP */
